/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Author
    Hrvoje Jasak, Wikki Ltd.  All rights reserved.
    Fethi Tekin, All rights reserved.
    Oliver Borm, All rights reserved.
    Gabriel Bousquet, All rights reserved.
\*---------------------------------------------------------------------------*/

#include "timeFunctionGgiFvMesh.H"
#include "Time.H"
#include "addToRunTimeSelectionTable.H"
#include "ZoneIDs.H"
// from subset
#include "cellSet.H"
#include "motionSolver.H"
#include "volFields.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(timeFunctionGgiFvMesh, 0);
    addToRunTimeSelectionTable(dynamicFvMesh, timeFunctionGgiFvMesh, IOobject);
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void Foam::timeFunctionGgiFvMesh::calcMovingPoints() const
{
    if (debug)
    {
        Info<< "void timeFunctionGgiFvMesh::calcMovingMasks() const : "
            << "Calculating point and cell masks"
            << endl;
    }

    if (movingPointsPtr_)
    {
        FatalErrorIn("void timeFunctionGgiFvMesh::calcMovingMasks() const")
            << "point mask already calculated"
            << abort(FatalError);
    }

    // Retrieve the cell zone Names
    const wordList cellZoneNames = cellZones().names();

    // Set the points
    movingPointsPtr_ = new vectorField(allPoints().size(), vector::zero);

    vectorField& movingPoints = *movingPointsPtr_;

    label movingPointsID = pointZones().findZoneID("insidePoints");

    const labelList& pointAddr = pointZones()[movingPointsID];
    
    forAll (pointAddr, pointI)
      {
        	movingPoints[pointAddr[pointI]] = vector(0, 180/Foam::mathematicalConstant::pi, 0);
      }

}
    
void Foam::timeFunctionGgiFvMesh::calcMovingRingMask() const
{
  Info <<"Entering calcMovingRingMask..."<<endl;
    // Set the point mask
    movingRingPointsMaskPtr_ = new scalarField(allPoints().size(), 0);
    scalarField& movingRingPointsMask = *movingRingPointsMaskPtr_;
    Info <<"About to make the moving ring mask..."<<endl;
    label movingRingPointsID = pointZones().findZoneID("outsideRing");

    const labelList& pointAddr = pointZones()[movingRingPointsID];
    
    forAll (pointAddr, pointI)
      {
        	movingRingPointsMask[pointAddr[pointI]] = 1;
      }
    Info <<"Moving Ring Mask done..."<<endl;
    
    Info <<"About to make the inside points mask..."<<endl;
    label insidePointsID = pointZones().findZoneID("insidePoints");

    const labelList& pointAddr2 = pointZones()[insidePointsID];
    
    forAll (pointAddr2, pointI)
      {
	movingRingPointsMask[pointAddr2[pointI]] = 1;
      }
    Info << "inside Points Mask done..."<<endl;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

// Construct from components
Foam::timeFunctionGgiFvMesh::timeFunctionGgiFvMesh
(
    const IOobject& io
)
:
    dynamicFvMesh(io),
    dict_
    (
        IOdictionary
        (
            IOobject
            (
                "dynamicMeshDict",
                time().constant(),
                *this,
                IOobject::MUST_READ,
                IOobject::AUTO_WRITE
            )
        ).subDict(typeName + "Coeffs")
    ),
    csPtr_
    (
        coordinateSystem::New
        (
            "coordinateSystem",
            dict_.subDict("coordinateSystem")
        )
    ),
    movingPointsPtr_(NULL),
    subsetMesh_
    (
        IOobject
        (
            "motion",
            io.time().constant(),
            *this,
            IOobject::NO_READ,
            IOobject::NO_WRITE
        ),
        *this
     ),
        motionPtr_(NULL)
{
  Info << "creating timeFunctionGgiFvMesh 1"<<endl;
    // from subset
    word setName = dict_.lookup("set");
    Info <<"About to set up the subset mesh... "<<setName<<endl;
    cellSet currentSet
    (
        IOobject
        (
            setName,
            io.time().constant(),
            polyMesh::meshSubDir/"sets",
            *this,
            IOobject::MUST_READ,
            IOobject::NO_WRITE
        )
    );
    
    subsetMesh_.setLargeCellSubset(currentSet, -1);

    // Create motion solver on the subset
    motionPtr_ = motionSolver::New(subsetMesh_.subMesh());
    Info<< "Subset mesh set"<<endl;
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::timeFunctionGgiFvMesh::~timeFunctionGgiFvMesh()
{
    deleteDemandDrivenData(movingPointsPtr_);
    deleteDemandDrivenData(movingRingPointsMaskPtr_);

}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

// Return moving points mask
const Foam::vectorField& Foam::timeFunctionGgiFvMesh::movingPoints() const
{
    if (!movingPointsPtr_)
    {
        calcMovingPoints();
    }

    return *movingPointsPtr_;
}

// Return moving points mask.  Moving points marked with 1
const Foam::scalarField& Foam::timeFunctionGgiFvMesh::movingRingPointsMask() const
{
    if (!movingRingPointsMaskPtr_)
    {
        calcMovingRingMask();
    }

    return *movingRingPointsMaskPtr_;
}

bool Foam::timeFunctionGgiFvMesh::update()
{
  
    // Get the points from the moving part
    pointField subsetPoints = motionPtr_->newPoints();

    //- Get copy of mesh points
    pointField p = points();

    //- Map the moving part
    const labelList& subsetPointAddr = subsetMesh_.pointMap();

    forAll (subsetPoints, subsetI)
    {
        p[subsetPointAddr[subsetI]] = subsetPoints[subsetI];
    }

    subsetMesh_.subMesh().movePoints(subsetPoints);

    //fvMesh::
    movePoints(p);
    vector dTransValue = 
      deplacement(time().time().value())
       - deplacement(time().time().value() - time().deltaTValue());
    movePoints
    (
        csPtr_->globalPosition
        (
            csPtr_->localPosition(allPoints())
	    + movingPoints()*
	    (pitch(time().time().value()) - pitch(time().time().value() - time().deltaTValue()))
	 )
	+ 
	dTransValue*movingRingPointsMask()
    );
    
    csPtr_->origin() += dTransValue;


    // The mesh is not morphing
    return false;
}


// from subset

const Foam::fvMeshSubset& Foam::timeFunctionGgiFvMesh::subsetMesh() const
{
    return subsetMesh_;
}


// ************************************************************************* //
